/*
 * Copyright (C) 2021  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef CREATE_CONTACT_STATE_FLAG_REQUEST_HH_BC58D5E7FE45E123888655AD302DFEDF//date "+%s.%N"|md5sum|tr "[a-f]" "[A-F]"
#define CREATE_CONTACT_STATE_FLAG_REQUEST_HH_BC58D5E7FE45E123888655AD302DFEDF

#include "src/impl/strong_type.hh"

#include "libpg/pg_rw_transaction.hh"

#include <boost/uuid/uuid.hpp>

namespace Fred {
namespace Identity {
namespace Impl {
namespace Lib {

struct ContactStateFlag
{
    template <typename> struct Type { };

    using ServerContactNameChangeProhibited = Type<struct ServerContactNameChangeProhibitedTag_>;
    using ServerContactOrganizationChangeProhibited = Type<struct ServerContactOrganizationChangeProhibitedTag_>;
    using ServerContactIdentChangeProhibited = Type<struct ServerContactIdentChangeProhibitedTag_>;
    using ServerContactPermanentAddressChangeProhibited = Type<struct ServerContactPermanentAddressChangeProhibitedTag_>;
};

using ContactUuid = StrongType<boost::uuids::uuid, struct ContactUuidTag_>;

template <typename T>
void create_contact_state_flag_request(
        const LibPg::PgRwTransaction& tx,
        const ContactUuid& contact_uuid,
        ContactStateFlag::Type<T> contact_state_flag);

std::string wrap_into_psql_representation(ContactStateFlag::ServerContactNameChangeProhibited);
std::string wrap_into_psql_representation(ContactStateFlag::ServerContactOrganizationChangeProhibited);
std::string wrap_into_psql_representation(ContactStateFlag::ServerContactIdentChangeProhibited);
std::string wrap_into_psql_representation(ContactStateFlag::ServerContactPermanentAddressChangeProhibited);

}//namespace Fred::Identity::Impl::Lib
}//namespace Fred::Identity::Impl
}//namespace Fred::Identity
}//namespace Fred

#endif//CREATE_CONTACT_STATE_FLAG_REQUEST_HH_BC58D5E7FE45E123888655AD302DFEDF
